/*
This file is part of mfaktc.
Copyright (C) 2012  George Woltman (woltman@alum.mit.edu)

mfaktc is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mfaktc is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mfaktc.  If not, see <http://www.gnu.org/licenses/>.
*/


__device__ static void mul_128_F64_95_initial_special(int128 *res, int128 a, int128 b)
/* res = a * b (only lower 128 bits of the result), a.d1, a.d2, a.d3 are zero, b.d1 is zero, b.d0 is one */
{
  asm("{\n\t"
      "mov.u32         %0, %4;\n\t"          /* (a.d0 * b.d0).lo */
      "mov.u32         %1, 0;\n\t"           /* (a.d0 * b.d1).lo */
      "mul.lo.u32      %2, %4, %10;\n\t"     /* (a.d0 * b.d2).lo */
      "mul.lo.u32      %3, %4, %11;\n\t"     /* (a.d0 * b.d3).lo */

      "mad.hi.u32      %3, %4, %10, %3;\n\t" /* (a.d0 * b.d2).hi */
      "}"
      : "=r" (res->d0), "=r" (res->d1), "=r" (res->d2), "=r" (res->d3)
      : "r" (a.d0), "r" (a.d1), "r" (a.d2), "r" (a.d3)
        "r" (b.d0), "r" (b.d1), "r" (b.d2), "r" (b.d3));
}


__device__ static void mul_128_F64_95_special(int128 *res, int128 a, int128 b)
/* res = a * b (only lower 128 bits of the result), b.d1 is zero, b.d0 is one */
{
  asm("{\n\t"
      "mov.u32         %0,          %4;\n\t" /*                    (a.d0 * b.d0).lo */
      "mov.u32         %1,          %5;\n\t" /*                    (a.d1 * b.d0).lo */
      "mad.lo.cc.u32   %2, %4, %10, %6;\n\t" /* (a.d0 * b.d2).lo + (a.d2 * b.d0).lo */
      "madc.lo.u32     %3, %4, %11, %7;\n\t" /* (a.d0 * b.d3).lo + (a.d3 * b.d0).lo */

      "mad.hi.u32      %3, %4, %10, %3;\n\t" /* (a.d0 * b.d2).hi */

      "mad.lo.u32      %3, %5, %10, %3;\n\t" /* (a.d1 * b.d2).lo */
      "}"
      : "=r" (res->d0), "=r" (res->d1), "=r" (res->d2), "=r" (res->d3)
      : "r" (a.d0), "r" (a.d1), "r" (a.d2), "r" (a.d3)
        "r" (b.d0), "r" (b.d1), "r" (b.d2), "r" (b.d3));
}


__device__ static void mul_160_128_F64_95_special(int160 *res, int160 a, int128 b)
/* res = a * b (only lower 160 bits of the result), b.d1 is zero, b.d0 is one */
{
  asm("{\n\t"
      "mov.u32         %0,          %5;\n\t" /*                    (a.d0 * b.d0).lo */
      "mov.u32         %1,          %6;\n\t" /*                    (a.d1 * b.d0).lo */
      "mad.lo.cc.u32   %2, %5, %12, %7;\n\t" /* (a.d0 * b.d2).lo + (a.d2 * b.d0).lo */
      "madc.lo.cc.u32  %3, %5, %13, %8;\n\t" /* (a.d0 * b.d3).lo + (a.d3 * b.d0).lo */
      "madc.lo.u32     %4, %7, %12, %9;\n\t" /* (a.d2 * b.d2).lo + (a.d4 * b.d0).lo */

      "mad.hi.cc.u32   %3, %5, %12, %3;\n\t" /* (a.d0 * b.d2).hi */
      "madc.hi.u32     %4, %5, %13, %4;\n\t" /* (a.d0 * b.d3).hi */

      "mad.lo.cc.u32   %3, %6, %12, %3;\n\t" /* (a.d1 * b.d2).lo */
      "madc.lo.u32     %4, %6, %13, %4;\n\t" /* (a.d1 * b.d3).lo */

      "mad.hi.u32      %4, %6, %12, %4;\n\t" /* (a.d1 * b.d2).hi */
      "}"
      : "=r" (res->d0), "=r" (res->d1), "=r" (res->d2), "=r" (res->d3), "=r" (res->d4)
      : "r" (a.d0), "r" (a.d1), "r" (a.d2), "r" (a.d3), "r" (a.d4)
        "r" (b.d0), "r" (b.d1), "r" (b.d2), "r" (b.d3));
}


__device__ static void mul_160_F64_95_initial_special(int160 *res, int160 a, int160 b)
/* res = a * b (only lower 160 bits of the result), a.d4 is zero, b.d1 is zero, b.d0 is one */
{
  asm("{\n\t"
      "mul.lo.u32      %4, %7, %12;\n\t"     /* (a.d2 * b.d2).lo */

      "mov.u32         %0,          %5;\n\t" /*                    (a.d0 * b.d0).lo */
      "mov.u32         %1,          %6;\n\t" /*                    (a.d1 * b.d0).lo */
      "mad.lo.cc.u32   %2, %5, %12, %7;\n\t" /* (a.d0 * b.d2).lo + (a.d2 * b.d0).lo */
      "madc.lo.cc.u32  %3, %5, %13, %8;\n\t" /* (a.d0 * b.d3).lo + (a.d3 * b.d0).lo */
      "madc.lo.u32     %4, %5, %14, %4;\n\t" /* (a.d0 * b.d4).lo */

      "mad.hi.cc.u32   %3, %5, %12, %3;\n\t" /* (a.d0 * b.d2).hi */
      "madc.hi.u32     %4, %5, %13, %4;\n\t" /* (a.d0 * b.d3).hi */

      "mad.lo.cc.u32   %3, %6, %12, %3;\n\t" /* (a.d1 * b.d2).lo */
      "madc.lo.u32     %4, %6, %13, %4;\n\t" /* (a.d1 * b.d3).lo */

      "mad.hi.u32      %4, %6, %12, %4;\n\t" /* (a.d1 * b.d2).hi */
      "}"
      : "=r" (res->d0), "=r" (res->d1), "=r" (res->d2), "=r" (res->d3), "=r" (res->d4)
      : "r" (a.d0), "r" (a.d1), "r" (a.d2), "r" (a.d3), "r" (a.d4)
        "r" (b.d0), "r" (b.d1), "r" (b.d2), "r" (b.d3), "r" (b.d4));
}


__device__ static void mul_160_F64_95_special(int160 *res, int160 a, int160 b)
/* res = a * b (only lower 160 bits of the result), b.d1 is zero, b.d0 is one */
{
  asm("{\n\t"
      "mov.u32         %0,          %5;\n\t" /*                    (a.d0 * b.d0).lo */
      "mov.u32         %1,          %6;\n\t" /*                    (a.d1 * b.d0).lo */
      "mad.lo.cc.u32   %2, %5, %12, %7;\n\t" /* (a.d0 * b.d2).lo + (a.d2 * b.d0).lo */
      "madc.lo.cc.u32  %3, %5, %13, %8;\n\t" /* (a.d0 * b.d3).lo + (a.d3 * b.d0).lo */
      "madc.lo.u32     %4, %5, %14, %9;\n\t" /* (a.d0 * b.d4).lo + (a.d4 * b.d0).lo */

      "mad.hi.cc.u32   %3, %5, %12, %3;\n\t" /* (a.d0 * b.d2).hi */
      "madc.hi.u32     %4, %5, %13, %4;\n\t" /* (a.d0 * b.d3).hi */

      "mad.lo.cc.u32   %3, %6, %12, %3;\n\t" /* (a.d1 * b.d2).lo */
      "madc.lo.u32     %4, %6, %13, %4;\n\t" /* (a.d1 * b.d3).lo */

      "mad.hi.u32      %4, %6, %12, %4;\n\t" /* (a.d1 * b.d2).hi */

      "mad.lo.u32      %4, %7, %12, %4;\n\t" /* (a.d2 * b.d2).lo */
      "}"
      : "=r" (res->d0), "=r" (res->d1), "=r" (res->d2), "=r" (res->d3), "=r" (res->d4)
      : "r" (a.d0), "r" (a.d1), "r" (a.d2), "r" (a.d3), "r" (a.d4)
        "r" (b.d0), "r" (b.d1), "r" (b.d2), "r" (b.d3), "r" (b.d4));
}


__device__ static void mul_192_160_F64_95_initial_special16(int192 *res, int192 a, int160 b)
/* res = a * b (only lower 192 bits of the result), a.d4, a,d5 are zero, b.d1 is zero, b.d0 is one */
{
  asm("{\n\t"
      "mul.lo.u32      %4, %8, %14;\n\t"      /* (a.d2 * b.d2).lo */
      "mul.lo.u32      %5, %8, %15;\n\t"      /* (a.d2 * b.d3).lo */

      "mov.u32         %0,          %6;\n\t"  /*                    (a.d0 * b.d0).lo */
      "mov.u32         %1,          %7;\n\t"  /*                    (a.d1 * b.d0).lo */
      "mad.lo.cc.u32   %2, %6, %14, %8;\n\t"  /* (a.d0 * b.d2).lo + (a.d2 * b.d0).lo */
      "madc.lo.cc.u32  %3, %6, %15, %9;\n\t"  /* (a.d0 * b.d3).lo + (a.d3 * b.d0).lo */
      "madc.lo.cc.u32  %4, %6, %16, %4;\n\t"  /* (a.d0 * b.d4).lo */
      "madc.lo.u32     %5, %9, %14, %5;\n\t"  /* (a.d3 * b.d2).lo */

      "mad.hi.cc.u32   %3, %6, %14, %3;\n\t"  /* (a.d0 * b.d2).hi */
      "madc.hi.cc.u32  %4, %6, %15, %4;\n\t"  /* (a.d0 * b.d3).hi */
      "madc.hi.u32     %5, %6, %16, %5;\n\t"  /* (a.d0 * b.d4).hi */

      "mad.lo.cc.u32   %3, %7, %14, %3;\n\t"  /* (a.d1 * b.d2).lo */
      "madc.lo.cc.u32  %4, %7, %15, %4;\n\t"  /* (a.d1 * b.d3).lo */
      "madc.lo.u32     %5, %7, %16, %5;\n\t"  /* (a.d1 * b.d4).lo */

      "mad.hi.cc.u32   %4, %7, %14, %4;\n\t"  /* (a.d1 * b.d2).hi */
      "madc.hi.u32     %5, %7, %15, %5;\n\t"  /* (a.d1 * b.d3).hi */

      "mad.hi.u32      %5, %8, %14, %5;\n\t"  /* (a.d2 * b.d2).hi */
      "}"
      : "=r" (res->d0), "=r" (res->d1), "=r" (res->d2), "=r" (res->d3), "=r" (res->d4), "=r" (res->d5)
      : "r" (a.d0), "r" (a.d1), "r" (a.d2), "r" (a.d3), "r" (a.d4), "r" (a.d5),
        "r" (b.d0), "r" (b.d1), "r" (b.d2), "r" (b.d3), "r" (b.d4));
}


__device__ static void mul_192_160_F64_95_special(int192 *res, int192 a, int160 b)
/* res = a * b (only lower 192 bits of the result), b.d1 is zero, b.d0 is one */
{
  asm("{\n\t"
      "mov.u32         %0,          %6;\n\t"  /*                    (a.d0 * b.d0).lo */
      "mov.u32         %1,          %7;\n\t"  /*                    (a.d1 * b.d0).lo */
      "mad.lo.cc.u32   %2, %6, %14, %8;\n\t"  /* (a.d0 * b.d2).lo + (a.d2 * b.d0).lo */
      "madc.lo.cc.u32  %3, %6, %15, %9;\n\t"  /* (a.d0 * b.d3).lo + (a.d3 * b.d0).lo */
      "madc.lo.cc.u32  %4, %6, %16, %10;\n\t" /* (a.d0 * b.d4).lo + (a.d4 * b.d0).lo */
      "madc.lo.u32     %5, %9, %14, %11;\n\t" /* (a.d3 * b.d2).lo + (a.d5 * b.d0).lo */

      "mad.hi.cc.u32   %3, %6, %14, %3;\n\t"  /* (a.d0 * b.d2).hi */
      "madc.hi.cc.u32  %4, %6, %15, %4;\n\t"  /* (a.d0 * b.d3).hi */
      "madc.hi.u32     %5, %6, %16, %5;\n\t"  /* (a.d0 * b.d4).hi */

      "mad.lo.cc.u32   %3, %7, %14, %3;\n\t"  /* (a.d1 * b.d2).lo */
      "madc.lo.cc.u32  %4, %7, %15, %4;\n\t"  /* (a.d1 * b.d3).lo */
      "madc.lo.u32     %5, %7, %16, %5;\n\t"  /* (a.d1 * b.d4).lo */

      "mad.hi.cc.u32   %4, %7, %14, %4;\n\t"  /* (a.d1 * b.d2).hi */
      "madc.hi.u32     %5, %7, %15, %5;\n\t"  /* (a.d1 * b.d3).hi */

      "mad.lo.cc.u32   %4, %8, %14, %4;\n\t"  /* (a.d2 * b.d2).lo */
      "madc.lo.u32     %5, %8, %15, %5;\n\t"  /* (a.d2 * b.d3).lo */

      "mad.hi.u32      %5, %8, %14, %5;\n\t"  /* (a.d2 * b.d2).hi */
      "}"
      : "=r" (res->d0), "=r" (res->d1), "=r" (res->d2), "=r" (res->d3), "=r" (res->d4), "=r" (res->d5)
      : "r" (a.d0), "r" (a.d1), "r" (a.d2), "r" (a.d3), "r" (a.d4), "r" (a.d5),
        "r" (b.d0), "r" (b.d1), "r" (b.d2), "r" (b.d3), "r" (b.d4));
}


//
// Kernel to factor Fermat numbers with factors f = k*2^N+1, 64 <= N <= 95.  Works on f between 97 and 108 bits inclusive
//

__global__ void
__launch_bounds__(THREADS_PER_BLOCK, KERNEL_MIN_BLOCKS) mfaktc_barrett108_F64_95gs(unsigned int exp, int96 k_base, unsigned int *bit_array, unsigned int bits_to_process, unsigned int *RES)
{
  int i, words_per_thread, sieve_word, k_bit_base, total_bit_count;
  int128 f, finalrem;
  __shared__ volatile unsigned short bitcount[256];	// Each thread of our block puts bit-counts here
  extern __shared__ unsigned short smem[];		// Write bits to test here.  Launching program must estimate
							// how much shared memory to allocate based on number of primes sieved.

  // Get pointer to section of the bit_array this thread is processing.

  words_per_thread = bits_to_process / 8192;
  bit_array += blockIdx.x * bits_to_process / 32 + threadIdx.x * words_per_thread;

// Count number of bits set in this thread's word(s) from the bit_array

  bitcount[threadIdx.x] = 0;
  for (i = 0; i < words_per_thread; i++)
    bitcount[threadIdx.x] += ___popcnt(bit_array[i]);

// Create total count of bits set in block up to and including this threads popcnt.
// Kudos to Rocke Verser for the population counting code.
// CAUTION:  Following requires 256 threads per block

  // First five tallies remain within one warp.  Should be in lock-step.
  if (threadIdx.x & 1)        // If we are running on any thread 0bxxxxxxx1, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[threadIdx.x - 1];

  if (threadIdx.x & 2)        // If we are running on any thread 0bxxxxxx1x, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 2) | 1];

  if (threadIdx.x & 4)        // If we are running on any thread 0bxxxxx1xx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 4) | 3];

  if (threadIdx.x & 8)        // If we are running on any thread 0bxxxx1xxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 8) | 7];

  if (threadIdx.x & 16)       // If we are running on any thread 0bxxx1xxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 16) | 15];

  // Further tallies are across warps.  Must synchronize
  __syncthreads();
  if (threadIdx.x  & 32)      // If we are running on any thread 0bxx1xxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 32) | 31];

  __syncthreads();
  if (threadIdx.x & 64)       // If we are running on any thread 0bx1xxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 64) | 63];

  __syncthreads();
  if (threadIdx.x & 128)       // If we are running on any thread 0b1xxxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[127];

  // At this point, bitcount[...] contains the total number of bits for the indexed
  // thread plus all lower-numbered threads.  I.e., bitcount[255] is the total count.

  __syncthreads();
  total_bit_count = bitcount[255];

//POSSIBLE OPTIMIZATION - bitcounts and smem could use the same memory space if we'd read bitcount into a register
// and sync threads before doing any writes to smem.

//POSSIBLE SANITY CHECK -- is there any way to test if total_bit_count exceeds the amount of shared memory allocated?

// Loop til this thread's section of the bit array is finished.

  sieve_word = *bit_array;
  k_bit_base = threadIdx.x * words_per_thread * 32;
  for (i = total_bit_count - bitcount[threadIdx.x]; ; i++) {
    int bit_to_test;

// Make sure we have a non-zero sieve word

    while (sieve_word == 0) {
      if (--words_per_thread == 0) break;
      sieve_word = *++bit_array;
      k_bit_base += 32;
    }

// Check if this thread has processed all its set bits

    if (sieve_word == 0) break;

// Find a bit to test in the sieve word

    bit_to_test = 31 - ___clz (sieve_word);
    sieve_word &= ~(1 << bit_to_test);

// Copy the k value to the shared memory array

    smem[i] = k_bit_base + bit_to_test;
  }

  __syncthreads();

// Here, all warps in our block have placed their candidates in shared memory.
// Now we can start TFing candidates.

// Pre-init values used to compute factor

  // Compute base k value
  k_base.d0 = __add_cc (k_base.d0, __umul32  (blockIdx.x * bits_to_process, NUM_CLASSES));
  k_base.d1 = __addc   (k_base.d1, __umul32hi(blockIdx.x * bits_to_process, NUM_CLASSES)); /* k values are limited to 64 bits */

// Loop til the k values written to shared memory are exhausted

  for (i = threadIdx.x; i < total_bit_count; i += THREADS_PER_BLOCK) {
    int96 k;
    int128 a, u, tmp128;
    int224 b, tmp224;
    float ff;
    int k_delta, shifter;

// Get the (k - k_base) value to test

    k_delta = smem[i];

// Compute f = k * 2^exp + 1

    k.d0 = __add_cc (k_base.d0, k_delta * NUM_CLASSES);	// k
    k.d1 = __addc   (k_base.d1, 0);

    f.d0 = 1;						// f = k * 2^exp + 1
    f.d1 = 0;
    f.d2 = (k.d0 << (exp - 64));
    f.d3 = (k.d1 << (exp - 64)) + (k.d0 >> (32 - (exp - 64)));

/*
ff = f as float
Precalculated here since it is the same for all steps in the following loop */
    ff= __uint2float_rn(f.d3);
    ff= ff * 4294967296.0f + __uint2float_rn(f.d2);
    ff=__int_as_float(0x3f7ffffb) / ff;			// just a little bit below 1.0f so we always underestimate the quotient

    tmp224.d6 = 0xFFFFFFFF;				// tmp224 is nearly 2^224
    tmp224.d5 = 0xFFFFFFFF; tmp224.d4 = 0xFFFFFFFF; tmp224.d3 = 0xFFFFFFFF;
    tmp224.d2 = 0xFFFFFFFF; tmp224.d1 = 0xFFFFFFFF; tmp224.d0 = 0xFFFFFFFF;

    // Could write optimized div_224_128 with so many tmp224 elements known to be zero
    div_224_128(&u,tmp224,f,ff);			// u = floor(2^224 / f).  This requires f >= 97 bits.

							// b_preinit = 2^128
							// a = b_preinit / 2^96 = 2^32
							// tmp256 = a * u = (b_preinit / 2^96) * (2^224 / f)     (ignore the floor functions for now)
    a.d0 = u.d3;					// a = tmp256 / 2^128, which if we do the math simplifies to the quotient: b_preinit / f

    mul_128_F64_95_initial_special(&tmp128, a, f);	// tmp128 = quotient * f, we only compute the low 128-bits here

    a.d0 = __sub_cc (0, tmp128.d0);			// Compute the remainder
    a.d1 = __subc_cc(0, tmp128.d1);			// we do not need the upper digits of b_preinit and tmp128 because the result is 0 after subtraction!
    a.d2 = __subc_cc(0, tmp128.d2);
    a.d3 = __subc   (0, tmp128.d3);

    for (shifter = 0; shifter < exp - 2 - 7; shifter++)
    {
							// On input a is at most 111.17 bits (see end of this loop)

      square_128_224(&b, a);				// b = a^2, b is at most 222.34 bits

      tmp128.d0 = b.d3;					// a = b / 2^96, a is at most 126.34 bits
      tmp128.d1 = b.d4;
      tmp128.d2 = b.d5;
      tmp128.d3 = b.d6;

      mul_128_256_no_low4(&a, tmp128, u);		// a = (b / 2^96) * (2^224 / f) / 2^128   (ignore the floor functions for now)
							// which if we do the math simplifies to the quotient: b / f

							// The quotient is off by at most 8.  A full mul_128_256 would add 7 partial results
							// into tmp256.d3 which could have generated 6 carries into tmp256.d4.
							// Also, since u was generated with the floor function, it could be low by up to
							// almost 1.  If we account for this a value up to a.d3 could have been added into
							// tmp256.d3 possibly generating a carry.  Similarly, a was generated by a floor
							// function, and could thus be low by almost 1.  If we account for this a value up
							// to u.d3 could have been added into tmp256.d3 possibly generating a carry.
							// A grand total of up to 8 carries lost.

      mul_128_F64_95_special(&tmp128, a, f);		// tmp128 = quotient * f, we only compute the low 128-bits here

      a.d0 = __sub_cc (b.d0, tmp128.d0);		// Compute the remainder
      a.d1 = __subc_cc(b.d1, tmp128.d1);		// we do not need the upper digits of b and tmp128 because the result is 0 after subtraction!
      a.d2 = __subc_cc(b.d2, tmp128.d2);
      a.d3 = __subc   (b.d3, tmp128.d3);
							// Since the quotient was up to 8 too small, the remainder has a maximum value of 9*f,
							// or 108 bits + log2 (9) bits, which is 111.17 bits.
    }

    mod_simple_128(&finalrem, a, f, ff);		// Adjustment.  The code above may produce an a that is too large by up to 8 times f.

#if 0
    if(cmp_ge_128(finalrem,f) && f.d3)
    {
      printf("EEEEEK, final finalrem is >= f\n, f: %X %X %X %X, a: %X %X %X %X", f.d3, f.d2, f.d1, f.d0, finalrem.d3, finalrem.d2, finalrem.d1, finalrem.d0);
    }
if (blockIdx.x == 0 && threadIdx.x == 4){
mul_128_256_no_low4(&tmp128,u,f);
printf ("    f: %08X%08X%08X%08X\r\n", f.d3, f.d2, f.d1, f.d0);
printf ("u    : %X %X %X %X\r\n", u.d3, u.d2, u.d1, u.d0);
printf ("u * f: %X %X %X %X\r\n", tmp128.d3, tmp128.d2, tmp128.d1, tmp128.d0);
printf ("  rem: %08X%08X%08X%08X\r\n", finalrem.d3, finalrem.d2, finalrem.d1, finalrem.d0);
}
#endif

/* check if we found a factor and write the factor to RES[] */

    if(finalrem.d1 == 0)
    if((finalrem.d2 == f.d2 && finalrem.d0 == 0 && finalrem.d3 == f.d3) ||
       (finalrem.d2 == 0    && finalrem.d0 == 1 && finalrem.d3 == 0))
    {
      int index=atomicInc(&RES[0],10000);
      if(index<10)				/* limit to 10 factors per class */
      {
        RES[index*6 + 1]=0;
	RES[index*6 + 2]=0;
	RES[index*6 + 3]=f.d3;
	RES[index*6 + 4]=f.d2;
	RES[index*6 + 5]=0;
	RES[index*6 + 6]=1;
      }
    }
  }

/* finally write occasional result for validation by C code */

  if(blockIdx.x == 0 && threadIdx.x == 0)
  {
    RES[84]=0;
    RES[85]=0;
    RES[86]=f.d3;
    RES[87]=f.d2;
    RES[88]=0;
    RES[89]=1;
    RES[94]=0;
    RES[95]=0;
    RES[96]=finalrem.d3;
    RES[97]=finalrem.d2;
    RES[98]=finalrem.d1;
    RES[99]=finalrem.d0;
  }
}


//
// Kernel to factor Fermat numbers with factors f = k*2^N+1, 64 <= N <= 95.  Works on f between 109 and 120 bits inclusive
//

__global__ void
__launch_bounds__(THREADS_PER_BLOCK, KERNEL_MIN_BLOCKS) mfaktc_barrett120_F64_95gs(unsigned int exp, int96 k_base, unsigned int *bit_array, unsigned int bits_to_process, unsigned int *RES, unsigned int bits_max)
{
  int i, words_per_thread, sieve_word, k_bit_base, total_bit_count;
  int128 f, finalrem;
  __shared__ volatile unsigned short bitcount[256];	// Each thread of our block puts bit-counts here
  extern __shared__ unsigned short smem[];		// Write bits to test here.  Launching program must estimate
							// how much shared memory to allocate based on number of primes sieved.

  // Get pointer to section of the bit_array this thread is processing.

  words_per_thread = bits_to_process / 8192;
  bit_array += blockIdx.x * bits_to_process / 32 + threadIdx.x * words_per_thread;

// Count number of bits set in this thread's word(s) from the bit_array

  bitcount[threadIdx.x] = 0;
  for (i = 0; i < words_per_thread; i++)
    bitcount[threadIdx.x] += ___popcnt(bit_array[i]);

// Create total count of bits set in block up to and including this threads popcnt.
// Kudos to Rocke Verser for the population counting code.
// CAUTION:  Following requires 256 threads per block

  // First five tallies remain within one warp.  Should be in lock-step.
  if (threadIdx.x & 1)        // If we are running on any thread 0bxxxxxxx1, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[threadIdx.x - 1];

  if (threadIdx.x & 2)        // If we are running on any thread 0bxxxxxx1x, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 2) | 1];

  if (threadIdx.x & 4)        // If we are running on any thread 0bxxxxx1xx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 4) | 3];

  if (threadIdx.x & 8)        // If we are running on any thread 0bxxxx1xxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 8) | 7];

  if (threadIdx.x & 16)       // If we are running on any thread 0bxxx1xxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 16) | 15];

  // Further tallies are across warps.  Must synchronize
  __syncthreads();
  if (threadIdx.x  & 32)      // If we are running on any thread 0bxx1xxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 32) | 31];

  __syncthreads();
  if (threadIdx.x & 64)       // If we are running on any thread 0bx1xxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 64) | 63];

  __syncthreads();
  if (threadIdx.x & 128)       // If we are running on any thread 0b1xxxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[127];

  // At this point, bitcount[...] contains the total number of bits for the indexed
  // thread plus all lower-numbered threads.  I.e., bitcount[255] is the total count.

  __syncthreads();
  total_bit_count = bitcount[255];

//POSSIBLE OPTIMIZATION - bitcounts and smem could use the same memory space if we'd read bitcount into a register
// and sync threads before doing any writes to smem.

//POSSIBLE SANITY CHECK -- is there any way to test if total_bit_count exceeds the amount of shared memory allocated?

// Loop til this thread's section of the bit array is finished.

  sieve_word = *bit_array;
  k_bit_base = threadIdx.x * words_per_thread * 32;
  for (i = total_bit_count - bitcount[threadIdx.x]; ; i++) {
    int bit_to_test;

// Make sure we have a non-zero sieve word

    while (sieve_word == 0) {
      if (--words_per_thread == 0) break;
      sieve_word = *++bit_array;
      k_bit_base += 32;
    }

// Check if this thread has processed all its set bits

    if (sieve_word == 0) break;

// Find a bit to test in the sieve word

    bit_to_test = 31 - ___clz (sieve_word);
    sieve_word &= ~(1 << bit_to_test);

// Copy the k value to the shared memory array

    smem[i] = k_bit_base + bit_to_test;
  }

  __syncthreads();

// Here, all warps in our block have placed their candidates in shared memory.
// Now we can start TFing candidates.

// Pre-init values used to compute factor

  // Compute base k value
  k_base.d0 = __add_cc (k_base.d0, __umul32  (blockIdx.x * bits_to_process, NUM_CLASSES));
  k_base.d1 = __addc   (k_base.d1, __umul32hi(blockIdx.x * bits_to_process, NUM_CLASSES)); /* k values are limited to 64 bits */

// Loop til the k values written to shared memory are exhausted

  for (i = threadIdx.x; i < total_bit_count; i += THREADS_PER_BLOCK) {
    int96 k;
    int128 a, u, tmp128;
    int256 b, tmp256;
    float ff;
    int k_delta, shifter;

// Get the (k - k_base) value to test

    k_delta = smem[i];

// Compute f = k * 2^exp + 1

    k.d0 = __add_cc (k_base.d0, k_delta * NUM_CLASSES);	// k
    k.d1 = __addc   (k_base.d1, 0);

    f.d0 = 1;						// f = k * 2^exp + 1
    f.d1 = 0;
    f.d2 = (k.d0 << (exp - 64));
    f.d3 = (k.d1 << (exp - 64)) + (k.d0 >> (32 - (exp - 64)));

/*
ff = f as float
Precalculated here since it is the same for all steps in the following loop */
    ff= __uint2float_rn(f.d3);
    ff= ff * 4294967296.0f + __uint2float_rn(f.d2);
    ff=__int_as_float(0x3f7ffffb) / ff;			// just a little bit below 1.0f so we always underestimate the quotient

    tmp256.d7 = 1 << (bits_max - 1);			// tmp256 = 2^(127 + bits_in_f)
    tmp256.d6 = 0; tmp256.d5 = 0; tmp256.d4 = 0;
    tmp256.d3 = 0; tmp256.d2 = 0; tmp256.d1 = 0; tmp256.d0 = 0;

    // Could write optimized div_256_128 with so many tmp256 elements known to be zero
    div_256_128(&u,tmp256,f,ff);			// u = floor(2^(127 + bits_in_f) / f), giving 128 bits of precision

							// b_preinit = 2^128
							// a = b_preinit / 2 ^ (bits_in_f - 1)
							// tmp256 = a * u = (b_preinit / 2 ^ (bits_in_f - 1)) * (2 ^ (127 + bits_in_f) / f)     (ignore the floor functions for now)
							// a = tmp256 / 2^128, which if we do the math simplifies to the quotient: b_preinit / f
    a.d0 = (u.d3 >> (bits_max - 1));

    mul_128_F64_95_initial_special(&tmp128, a, f);	// tmp128 = quotient * f, we only compute the low 128-bits here

    a.d0 = __sub_cc (0, tmp128.d0);			// Compute the remainder
    a.d1 = __subc_cc(0, tmp128.d1);			// we do not need the upper digits of b_preinit and tmp128 because the result is 0 after subtraction!
    a.d2 = __subc_cc(0, tmp128.d2);
    a.d3 = __subc   (0, tmp128.d3);

    for (shifter = 0; shifter < exp - 2 - 7; shifter++)
    {
							// On input a is at most 123.17 bits (see end of this loop)

      square_128_256(&b, a);				// b = a^2, b is at most 246.34 bits

      tmp128.d0 = (b.d3 >> (bits_max - 1)) + (b.d4 << (32 - (bits_max - 1))); // a = b / (2 ^ (bits_in_f - 1)), a is at most 127.34 bits
      tmp128.d1 = (b.d4 >> (bits_max - 1)) + (b.d5 << (32 - (bits_max - 1)));
      tmp128.d2 = (b.d5 >> (bits_max - 1)) + (b.d6 << (32 - (bits_max - 1)));
      tmp128.d3 = (b.d6 >> (bits_max - 1)) + (b.d7 << (32 - (bits_max - 1)));

      mul_128_256_no_low4(&a, tmp128, u);		// a = (b / 2 ^ (bits_in_f - 1)) * (2 ^ (127 + bits_in_f) / f) / 2^128  (ignore the floor functions for now)
							// which if we do the math simplifies to the quotient: b / f

							// The quotient is off by at most 8.  A full mul_128_256 would add 7 partial results
							// into tmp256.d3 which could have generated 6 carries into tmp256.d4.
							// Also, since u was generated with the floor function, it could be low by up to
							// almost 1.  If we account for this a value up to a.d3 could have been added into
							// tmp256.d3 possibly generating a carry.  Similarly, a was generated by a floor
							// function, and could thus be low by almost 1.  If we account for this a value up
							// to u.d3 could have been added into tmp256.d3 possibly generating a carry.
							// A grand total of up to 8 carries lost.

      mul_128_F64_95_special(&tmp128, a, f);		// tmp128 = quotient * f, we only compute the low 128-bits here

      a.d0 = __sub_cc (b.d0, tmp128.d0);		// Compute the remainder
      a.d1 = __subc_cc(b.d1, tmp128.d1);		// we do not need the upper digits of b and tmp160 because the result is 0 after subtraction!
      a.d2 = __subc_cc(b.d2, tmp128.d2);
      a.d3 = __subc   (b.d3, tmp128.d3);
							// Since the quotient was up to 8 too small, the remainder has a maximum value of 9*f,
							// or 120 bits + log2 (9) bits, which is 123.17 bits.
    }

    mod_simple_128(&finalrem, a, f, ff);		// Adjustment.  The code above may produce an a that is too large by up to 8 times f.

#if 0
    if(cmp_ge_128(finalrem,f))
    {
      printf("EEEEEK, final finalrem is >= f\n");
    }
if (blockIdx.x == 0 && threadIdx.x == 4){
mul_128_256_no_low4(&tmp128,u,f);
printf ("    f: %08X%08X%08X%08X\r\n", f.d3, f.d2, f.d1, f.d0);
printf ("u    : %X %X %X %X\r\n", u.d3, u.d2, u.d1, u.d0);
printf ("u * f: %X %X %X %X\r\n", tmp128.d3, tmp128.d2, tmp128.d1, tmp128.d0);
printf ("  rem: %08X%08X%08X%08X\r\n", finalrem.d3, finalrem.d2, finalrem.d1, finalrem.d0);
}
#endif

/* check if we found a factor and write the factor to RES[] */

    if(finalrem.d1 == 0)
    if((finalrem.d2 == f.d2 && finalrem.d0 == 0 && finalrem.d3 == f.d3) ||
       (finalrem.d2 == 0    && finalrem.d0 == 1 && finalrem.d3 == 0))
    {
      int index=atomicInc(&RES[0],10000);
      if(index<10)				/* limit to 10 factors per class */
      {
        RES[index*6 + 1]=0;
	RES[index*6 + 2]=0;
	RES[index*6 + 3]=f.d3;
	RES[index*6 + 4]=f.d2;
	RES[index*6 + 5]=0;
	RES[index*6 + 6]=1;
      }
    }
  }

/* finally write occasional result for validation by C code */

  if(blockIdx.x == 0 && threadIdx.x == 0)
  {
    RES[84]=0;
    RES[85]=0;
    RES[86]=f.d3;
    RES[87]=f.d2;
    RES[88]=0;
    RES[89]=1;
    RES[94]=0;
    RES[95]=0;
    RES[96]=finalrem.d3;
    RES[97]=finalrem.d2;
    RES[98]=finalrem.d1;
    RES[99]=finalrem.d0;
  }
}


//
// Kernel to factor Fermat numbers with factors f = k*2^N+1, 64 <= N <= 95.  Works on f between 121 and 128 bits inclusive
//

__global__ void
__launch_bounds__(THREADS_PER_BLOCK, KERNEL_MIN_BLOCKS) mfaktc_barrett128_F64_95gs(unsigned int exp, int96 k_base, unsigned int *bit_array, unsigned int bits_to_process, unsigned int *RES)
{
  int i, words_per_thread, sieve_word, k_bit_base, total_bit_count;
  int128 f, finalrem;
  __shared__ volatile unsigned short bitcount[256];	// Each thread of our block puts bit-counts here
  extern __shared__ unsigned short smem[];		// Write bits to test here.  Launching program must estimate
							// how much shared memory to allocate based on number of primes sieved.

  // Get pointer to section of the bit_array this thread is processing.

  words_per_thread = bits_to_process / 8192;
  bit_array += blockIdx.x * bits_to_process / 32 + threadIdx.x * words_per_thread;

// Count number of bits set in this thread's word(s) from the bit_array

  bitcount[threadIdx.x] = 0;
  for (i = 0; i < words_per_thread; i++)
    bitcount[threadIdx.x] += ___popcnt(bit_array[i]);

// Create total count of bits set in block up to and including this threads popcnt.
// Kudos to Rocke Verser for the population counting code.
// CAUTION:  Following requires 256 threads per block

  // First five tallies remain within one warp.  Should be in lock-step.
  if (threadIdx.x & 1)        // If we are running on any thread 0bxxxxxxx1, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[threadIdx.x - 1];

  if (threadIdx.x & 2)        // If we are running on any thread 0bxxxxxx1x, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 2) | 1];

  if (threadIdx.x & 4)        // If we are running on any thread 0bxxxxx1xx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 4) | 3];

  if (threadIdx.x & 8)        // If we are running on any thread 0bxxxx1xxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 8) | 7];

  if (threadIdx.x & 16)       // If we are running on any thread 0bxxx1xxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 16) | 15];

  // Further tallies are across warps.  Must synchronize
  __syncthreads();
  if (threadIdx.x  & 32)      // If we are running on any thread 0bxx1xxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 32) | 31];

  __syncthreads();
  if (threadIdx.x & 64)       // If we are running on any thread 0bx1xxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 64) | 63];

  __syncthreads();
  if (threadIdx.x & 128)       // If we are running on any thread 0b1xxxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[127];

  // At this point, bitcount[...] contains the total number of bits for the indexed
  // thread plus all lower-numbered threads.  I.e., bitcount[255] is the total count.

  __syncthreads();
  total_bit_count = bitcount[255];

//POSSIBLE OPTIMIZATION - bitcounts and smem could use the same memory space if we'd read bitcount into a register
// and sync threads before doing any writes to smem.

//POSSIBLE SANITY CHECK -- is there any way to test if total_bit_count exceeds the amount of shared memory allocated?

// Loop til this thread's section of the bit array is finished.

  sieve_word = *bit_array;
  k_bit_base = threadIdx.x * words_per_thread * 32;
  for (i = total_bit_count - bitcount[threadIdx.x]; ; i++) {
    int bit_to_test;

// Make sure we have a non-zero sieve word

    while (sieve_word == 0) {
      if (--words_per_thread == 0) break;
      sieve_word = *++bit_array;
      k_bit_base += 32;
    }

// Check if this thread has processed all its set bits

    if (sieve_word == 0) break;

// Find a bit to test in the sieve word

    bit_to_test = 31 - ___clz (sieve_word);
    sieve_word &= ~(1 << bit_to_test);

// Copy the k value to the shared memory array

    smem[i] = k_bit_base + bit_to_test;
  }

  __syncthreads();

// Here, all warps in our block have placed their candidates in shared memory.
// Now we can start TFing candidates.

// Pre-init values used to compute factor

  // Compute base k value
  k_base.d0 = __add_cc (k_base.d0, __umul32  (blockIdx.x * bits_to_process, NUM_CLASSES));
  k_base.d1 = __addc   (k_base.d1, __umul32hi(blockIdx.x * bits_to_process, NUM_CLASSES)); /* k values are limited to 64 bits */

// Loop til the k values written to shared memory are exhausted

  for (i = threadIdx.x; i < total_bit_count; i += THREADS_PER_BLOCK) {
    int96 k;
    int160 a, u, tmp160;
    int288 b, tmp288;
    float ff;
    int k_delta, shifter;

// Get the (k - k_base) value to test

    k_delta = smem[i];

// Compute f = k * 2^exp + 1

    k.d0 = __add_cc (k_base.d0, k_delta * NUM_CLASSES);	// k
    k.d1 = __addc   (k_base.d1, 0);

    f.d0 = 1;						// f = k * 2^exp + 1
    f.d1 = 0;
    f.d2 = (k.d0 << (exp - 64));
    f.d3 = (k.d1 << (exp - 64)) + (k.d0 >> (32 - (exp - 64)));

/*
ff = f as float
Precalculated here since it is the same for all steps in the following loop */
    ff= __uint2float_rn(f.d3);
    ff= ff * 4294967296.0f + __uint2float_rn(f.d2);
    ff=__int_as_float(0x3f7ffffb) / ff;			// just a little bit below 1.0f so we always underestimate the quotient

    tmp288.d8 = 0x10000;				// tmp288 is 2^272
    tmp288.d7 = 0; tmp288.d6 = 0; tmp288.d5 = 0; tmp288.d4 = 0;
    tmp288.d3 = 0; tmp288.d2 = 0; tmp288.d1 = 0; tmp288.d0 = 0;

    // Could write optimized div_288_128 with so many tmp288 elements known to be zero
    div_288_128(&u,tmp288,f,ff);			// u = floor(2^272 / f).  This requires f >= 113 bits.

							// b_preinit = 2^256
							// a = b_preinit / 2^112 = 2^144
							// tmp320 = a * u = (b_preinit / 2^112) * (2^272 / f)     (ignore the floor functions for now)
    a.d0 = (u.d1 << 16) + (u.d0 >> 16);			// a = tmp320 / 2^160, which if we do the math simplifies to the quotient: b_preinit / f
    a.d1 = (u.d2 << 16) + (u.d1 >> 16);
    a.d2 = (u.d3 << 16) + (u.d2 >> 16);
    a.d3 = (u.d4 << 16) + (u.d3 >> 16);
    a.d4 =                (u.d4 >> 16);

    mul_160_128_F64_95_special(&tmp160, a, f);		// tmp160 = quotient * f, we only compute the low 160-bits here

    a.d0 = __sub_cc (0, tmp160.d0);			// Compute the remainder
    a.d1 = __subc_cc(0, tmp160.d1);			// we do not need the upper digits of b_preinit and tmp192 because the result is 0 after subtraction!
    a.d2 = __subc_cc(0, tmp160.d2);
    a.d3 = __subc_cc(0, tmp160.d3);
    a.d4 = __subc   (0, tmp160.d4);

    for (shifter = 0; shifter < exp - 2 - 8; shifter++)
    {
							// On input a is at most 131.459 bits (see end of this loop)

      square_160_288(&b, a);				// b = a^2, b is at most 262.918 bits

      tmp160.d0 = (b.d4 << 16) + (b.d3 >> 16);		// a = b / 2^112, a is at most 150.918 bits
      tmp160.d1 = (b.d5 << 16) + (b.d4 >> 16);
      tmp160.d2 = (b.d6 << 16) + (b.d5 >> 16);
      tmp160.d3 = (b.d7 << 16) + (b.d6 >> 16);
      tmp160.d4 = (b.d8 << 16) + (b.d7 >> 16);

      mul_160_320_no_low5(&a, tmp160, u);		// a = (b / 2^112) * (2^272 / f) / 2^160    (ignore the floor functions for now)
							// which if we do the math simplifies to the quotient: b / f

							// The quotient is off by at most 10.  A full mul_160_320 would add 9 partial results
							// into tmp320.d4 which could have generated 8 carries into tmp320.d5.
							// Also, since u was generated with the floor function, it could be low by up to
							// almost 1.  If we account for this a value up to a.d4 could have been added into
							// tmp320.d4 possibly generating a carry.  Similarly, a was generated by a floor
							// function, and could thus be low by almost 1.  If we account for this a value up
							// to u.d4 could have been added into tmp320.d4 possibly generating a carry.
							// A grand total of up to 10 carries lost.

      mul_160_128_F64_95_special(&tmp160, a, f);	// tmp160 = quotient * f, we only compute the low 160-bits here

      a.d0 = __sub_cc (b.d0, tmp160.d0);		// Compute the remainder
      a.d1 = __subc_cc(b.d1, tmp160.d1);		// we do not need the upper digits of b and tmp160 because the result is 0 after subtraction!
      a.d2 = __subc_cc(b.d2, tmp160.d2);
      a.d3 = __subc_cc(b.d3, tmp160.d3);
      a.d4 = __subc   (b.d4, tmp160.d4);
							// Since the quotient was up to 10 too small, the remainder has a maximum value of 11*f,
							// or 128 bits + log2 (11) bits, which is 131.459 bits.
    }

    mod_simple_160_128(&finalrem, a, f, ff);		// Adjustment.  The code above may produce an a that is too large by up to 11 times f.

#if 0
    if(cmp_ge_128(finalrem,f) && (f.d3 & 0xFFFF0000))
    {
      printf("EEEEEK, final rem is >= f\n");
    }
if ((blockIdx.x == 0 && threadIdx.x == 4)){
int160 f160;	    
f160.d0 = f.d0;
f160.d1 = f.d1;
f160.d2 = f.d2;
f160.d3 = f.d3;
f160.d4 = 0;
mul_160_320_no_low5(&tmp160,u,f160);
printf ("    f: %08X%08X%08X%08X\r\n", f.d3, f.d2, f.d1, f.d0);
printf ("u    : %X %X %X %X %X\r\n", u.d4, u.d3, u.d2, u.d1, u.d0);
printf ("u * f: %X %X %X %X %X\r\n", tmp160.d4, tmp160.d3, tmp160.d2, tmp160.d1, tmp160.d0);
printf ("  rem: %08X%08X%08X%08X\r\n", finalrem.d3, finalrem.d2, finalrem.d1, finalrem.d0);
}
#endif

/* check if we found a factor and write the factor to RES[] */

    if(finalrem.d1 == 0)
    if((finalrem.d2 == f.d2 && finalrem.d0 == 0 && finalrem.d3 == f.d3) ||
       (finalrem.d2 == 0    && finalrem.d0 == 1 && finalrem.d3 == 0))
    {
      int index=atomicInc(&RES[0],10000);
      if(index<10)				/* limit to 10 factors per class */
      {
        RES[index*6 + 1]=0;
	RES[index*6 + 2]=0;
	RES[index*6 + 3]=f.d3;
	RES[index*6 + 4]=f.d2;
	RES[index*6 + 5]=0;
	RES[index*6 + 6]=1;
      }
    }
  }

/* finally write occasional result for validation by C code */

  if(blockIdx.x == 0 && threadIdx.x == 0)
  {
    RES[84]=0;
    RES[85]=0;
    RES[86]=f.d3;
    RES[87]=f.d2;
    RES[88]=0;
    RES[89]=1;
    RES[94]=0;
    RES[95]=0;
    RES[96]=finalrem.d3;
    RES[97]=finalrem.d2;
    RES[98]=finalrem.d1;
    RES[99]=finalrem.d0;
  }
}

//
// Kernel to factor Fermat numbers with factors f = k*2^N+1, 64 <= N <= 95.  Works on f between 129 and 140 bits inclusive
//

__global__ void
__launch_bounds__(THREADS_PER_BLOCK, KERNEL_MIN_BLOCKS) mfaktc_barrett140_F64_95gs(unsigned int exp, int96 k_base, unsigned int *bit_array, unsigned int bits_to_process, unsigned int *RES)
{
  int i, words_per_thread, sieve_word, k_bit_base, total_bit_count;
  int160 f, finalrem;
  __shared__ volatile unsigned short bitcount[256];	// Each thread of our block puts bit-counts here
  extern __shared__ unsigned short smem[];		// Write bits to test here.  Launching program must estimate
							// how much shared memory to allocate based on number of primes sieved.

  // Get pointer to section of the bit_array this thread is processing.

  words_per_thread = bits_to_process / 8192;
  bit_array += blockIdx.x * bits_to_process / 32 + threadIdx.x * words_per_thread;

// Count number of bits set in this thread's word(s) from the bit_array

  bitcount[threadIdx.x] = 0;
  for (i = 0; i < words_per_thread; i++)
    bitcount[threadIdx.x] += ___popcnt(bit_array[i]);

// Create total count of bits set in block up to and including this threads popcnt.
// Kudos to Rocke Verser for the population counting code.
// CAUTION:  Following requires 256 threads per block

  // First five tallies remain within one warp.  Should be in lock-step.
  if (threadIdx.x & 1)        // If we are running on any thread 0bxxxxxxx1, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[threadIdx.x - 1];

  if (threadIdx.x & 2)        // If we are running on any thread 0bxxxxxx1x, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 2) | 1];

  if (threadIdx.x & 4)        // If we are running on any thread 0bxxxxx1xx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 4) | 3];

  if (threadIdx.x & 8)        // If we are running on any thread 0bxxxx1xxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 8) | 7];

  if (threadIdx.x & 16)       // If we are running on any thread 0bxxx1xxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 16) | 15];

  // Further tallies are across warps.  Must synchronize
  __syncthreads();
  if (threadIdx.x  & 32)      // If we are running on any thread 0bxx1xxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 32) | 31];

  __syncthreads();
  if (threadIdx.x & 64)       // If we are running on any thread 0bx1xxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 64) | 63];

  __syncthreads();
  if (threadIdx.x & 128)       // If we are running on any thread 0b1xxxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[127];

  // At this point, bitcount[...] contains the total number of bits for the indexed
  // thread plus all lower-numbered threads.  I.e., bitcount[255] is the total count.

  __syncthreads();
  total_bit_count = bitcount[255];

//POSSIBLE OPTIMIZATION - bitcounts and smem could use the same memory space if we'd read bitcount into a register
// and sync threads before doing any writes to smem.

//POSSIBLE SANITY CHECK -- is there any way to test if total_bit_count exceeds the amount of shared memory allocated?

// Loop til this thread's section of the bit array is finished.

  sieve_word = *bit_array;
  k_bit_base = threadIdx.x * words_per_thread * 32;
  for (i = total_bit_count - bitcount[threadIdx.x]; ; i++) {
    int bit_to_test;

// Make sure we have a non-zero sieve word

    while (sieve_word == 0) {
      if (--words_per_thread == 0) break;
      sieve_word = *++bit_array;
      k_bit_base += 32;
    }

// Check if this thread has processed all its set bits

    if (sieve_word == 0) break;

// Find a bit to test in the sieve word

    bit_to_test = 31 - ___clz (sieve_word);
    sieve_word &= ~(1 << bit_to_test);

// Copy the k value to the shared memory array

    smem[i] = k_bit_base + bit_to_test;
  }

  __syncthreads();

// Here, all warps in our block have placed their candidates in shared memory.
// Now we can start TFing candidates.

// Pre-init values used to compute factor

  // Compute base k value
  k_base.d0 = __add_cc (k_base.d0, __umul32  (blockIdx.x * bits_to_process, NUM_CLASSES));
  k_base.d1 = __addc   (k_base.d1, __umul32hi(blockIdx.x * bits_to_process, NUM_CLASSES)); /* k values are limited to 64 bits */

// Loop til the k values written to shared memory are exhausted

  for (i = threadIdx.x; i < total_bit_count; i += THREADS_PER_BLOCK) {
    int96 k;
    int160 a, u, tmp160;
    int288 b, tmp288;
    float ff;
    int k_delta, shifter;

// Get the (k - k_base) value to test

    k_delta = smem[i];

// Compute f = k * 2^exp + 1

    k.d0 = __add_cc (k_base.d0, k_delta * NUM_CLASSES);	// k
    k.d1 = __addc   (k_base.d1, 0);

    f.d0 = 1;						// f = k * 2^exp + 1
    f.d1 = 0;
    f.d2 = (k.d0 << (exp - 64));
    f.d3 = (k.d1 << (exp - 64)) + (k.d0 >> (32 - (exp - 64)));
    f.d4 =                        (k.d1 >> (32 - (exp - 64)));

/*
ff = f as float
Precalculated here since it is the same for all steps in the following loop */
    ff= __uint2float_rn(f.d4);
    ff= ff * 4294967296.0f + __uint2float_rn(f.d3);
    ff=__int_as_float(0x3f7ffffb) / ff;			// just a little bit below 1.0f so we always underestimate the quotient

    tmp288.d8 = 0xFFFFFFFF;				// tmp288 is nearly 2^288
    tmp288.d7 = 0xFFFFFFFF; tmp288.d6 = 0xFFFFFFFF; tmp288.d5 = 0xFFFFFFFF; tmp288.d4 = 0xFFFFFFFF;
    tmp288.d3 = 0xFFFFFFFF; tmp288.d2 = 0xFFFFFFFF; tmp288.d1 = 0xFFFFFFFF; tmp288.d0 = 0xFFFFFFFF;

    // Could write optimized div_288_160 with so many tmp288 elements known to be zero
    div_288_160(&u,tmp288,f,ff);			// u = floor(2^288 / f).  This requires f >= 129 bits.

							// b_preinit = 2^256
							// a = b_preinit / 2^128 = 2^144
							// tmp320 = a * u = (b_preinit / 2^128) * (2^288 / f)     (ignore the floor functions for now)
    a.d0 = u.d1;					// a = tmp320 / 2^160, which if we do the math simplifies to the quotient: b_preinit / f
    a.d1 = u.d2;
    a.d2 = u.d3;
    a.d3 = u.d4;

    mul_160_F64_95_initial_special(&tmp160, a, f);	// tmp160 = quotient * f, we only compute the low 160-bits here

    a.d0 = __sub_cc (0, tmp160.d0);			// Compute the remainder
    a.d1 = __subc_cc(0, tmp160.d1);			// we do not need the upper digits of b_preinit and tmp192 because the result is 0 after subtraction!
    a.d2 = __subc_cc(0, tmp160.d2);
    a.d3 = __subc_cc(0, tmp160.d3);
    a.d4 = __subc   (0, tmp160.d4);

    for (shifter = 0; shifter < exp - 2 - 8; shifter++)
    {
							// On input a is at most 143.459 bits (see end of this loop)

      square_160_288(&b, a);				// b = a^2, b is at most 286.918 bits

      tmp160.d0 = b.d4;					// a = b / 2^128, a is at most 158.918 bits
      tmp160.d1 = b.d5;
      tmp160.d2 = b.d6;
      tmp160.d3 = b.d7;
      tmp160.d4 = b.d8;

      mul_160_320_no_low5(&a, tmp160, u);		// a = (b / 2^128) * (2^288 / f) / 2^160    (ignore the floor functions for now)
							// which if we do the math simplifies to the quotient: b / f

							// The quotient is off by at most 10.  A full mul_160_320 would add 9 partial results
							// into tmp320.d4 which could have generated 8 carries into tmp320.d5.
							// Also, since u was generated with the floor function, it could be low by up to
							// almost 1.  If we account for this a value up to a.d4 could have been added into
							// tmp320.d4 possibly generating a carry.  Similarly, a was generated by a floor
							// function, and could thus be low by almost 1.  If we account for this a value up
							// to u.d4 could have been added into tmp320.d4 possibly generating a carry.
							// A grand total of up to 10 carries lost.

      mul_160_F64_95_special(&tmp160, a, f);		// tmp160 = quotient * f, we only compute the low 160-bits here

      a.d0 = __sub_cc (b.d0, tmp160.d0);		// Compute the remainder
      a.d1 = __subc_cc(b.d1, tmp160.d1);		// we do not need the upper digits of b and tmp160 because the result is 0 after subtraction!
      a.d2 = __subc_cc(b.d2, tmp160.d2);
      a.d3 = __subc_cc(b.d3, tmp160.d3);
      a.d4 = __subc   (b.d4, tmp160.d4);
							// Since the quotient was up to 10 too small, the remainder has a maximum value of 11*f,
							// or 140 bits + log2 (11) bits, which is 143.459 bits.
    }

    mod_simple_160(&finalrem, a, f, ff);		// Adjustment.  The code above may produce an a that is too large by up to 10 times f.

#if 0
    if(cmp_ge_160(finalrem,f) && f.d4)
    {
      printf("EEEEEK, final rem is >= f\n");
    }
if (blockIdx.x == 12 && threadIdx.x == 4){
mul_160_320_no_low5(&tmp160,u,f);
printf ("    f: %08X%08X%08X%08X%08X\r\n", f.d4, f.d3, f.d2, f.d1, f.d0);
printf ("u    : %X %X %X %X %X\r\n", u.d4, u.d3, u.d2, u.d1, u.d0);
printf ("u * f: %X %X %X %X %X\r\n", tmp160.d4, tmp160.d3, tmp160.d2, tmp160.d1, tmp160.d0);
printf ("  rem: %08X%08X%08X%08X%08X\r\n", finalrem.d4, finalrem.d3, finalrem.d2, finalrem.d1, finalrem.d0);
}
#endif

/* check if we found a factor and write the factor to RES[] */

    if(finalrem.d1 == 0)
    if((finalrem.d3 == f.d3 && finalrem.d2 == f.d2 && finalrem.d0 == 0 && finalrem.d4 == f.d4) ||
       (finalrem.d3 == 0    && finalrem.d2 == 0    && finalrem.d0 == 1 && finalrem.d4 == 0))
    {
      int index=atomicInc(&RES[0],10000);
      if(index<10)				/* limit to 10 factors per class */
      {
        RES[index*6 + 1]=0;
	RES[index*6 + 2]=f.d4;
	RES[index*6 + 3]=f.d3;
	RES[index*6 + 4]=f.d2;
	RES[index*6 + 5]=0;
	RES[index*6 + 6]=1;
      }
    }
  }

/* finally write occasional result for validation by C code */

  if(blockIdx.x == 0 && threadIdx.x == 0)
  {
    RES[84]=0;
    RES[85]=f.d4;
    RES[86]=f.d3;
    RES[87]=f.d2;
    RES[88]=0;
    RES[89]=1;
    RES[94]=0;
    RES[95]=finalrem.d4;
    RES[96]=finalrem.d3;
    RES[97]=finalrem.d2;
    RES[98]=finalrem.d1;
    RES[99]=finalrem.d0;
  }
}


//
// Kernel to factor Fermat numbers with factors f = k*2^N+1, 64 <= N <= 95.  Works on f between 141 and 152 bits inclusive
//

__global__ void
__launch_bounds__(THREADS_PER_BLOCK, KERNEL_MIN_BLOCKS) mfaktc_barrett152_F64_95gs(unsigned int exp, int96 k_base, unsigned int *bit_array, unsigned int bits_to_process, unsigned int *RES, unsigned int bits_max)
{
  int i, words_per_thread, sieve_word, k_bit_base, total_bit_count;
  int160 f, finalrem;
  __shared__ volatile unsigned short bitcount[256];	// Each thread of our block puts bit-counts here
  extern __shared__ unsigned short smem[];		// Write bits to test here.  Launching program must estimate
							// how much shared memory to allocate based on number of primes sieved.

  // Get pointer to section of the bit_array this thread is processing.

  words_per_thread = bits_to_process / 8192;
  bit_array += blockIdx.x * bits_to_process / 32 + threadIdx.x * words_per_thread;

// Count number of bits set in this thread's word(s) from the bit_array

  bitcount[threadIdx.x] = 0;
  for (i = 0; i < words_per_thread; i++)
    bitcount[threadIdx.x] += ___popcnt(bit_array[i]);

// Create total count of bits set in block up to and including this threads popcnt.
// Kudos to Rocke Verser for the population counting code.
// CAUTION:  Following requires 256 threads per block

  // First five tallies remain within one warp.  Should be in lock-step.
  if (threadIdx.x & 1)        // If we are running on any thread 0bxxxxxxx1, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[threadIdx.x - 1];

  if (threadIdx.x & 2)        // If we are running on any thread 0bxxxxxx1x, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 2) | 1];

  if (threadIdx.x & 4)        // If we are running on any thread 0bxxxxx1xx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 4) | 3];

  if (threadIdx.x & 8)        // If we are running on any thread 0bxxxx1xxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 8) | 7];

  if (threadIdx.x & 16)       // If we are running on any thread 0bxxx1xxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 16) | 15];

  // Further tallies are across warps.  Must synchronize
  __syncthreads();
  if (threadIdx.x  & 32)      // If we are running on any thread 0bxx1xxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 32) | 31];

  __syncthreads();
  if (threadIdx.x & 64)       // If we are running on any thread 0bx1xxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 64) | 63];

  __syncthreads();
  if (threadIdx.x & 128)       // If we are running on any thread 0b1xxxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[127];

  // At this point, bitcount[...] contains the total number of bits for the indexed
  // thread plus all lower-numbered threads.  I.e., bitcount[255] is the total count.

  __syncthreads();
  total_bit_count = bitcount[255];

//POSSIBLE OPTIMIZATION - bitcounts and smem could use the same memory space if we'd read bitcount into a register
// and sync threads before doing any writes to smem.

//POSSIBLE SANITY CHECK -- is there any way to test if total_bit_count exceeds the amount of shared memory allocated?

// Loop til this thread's section of the bit array is finished.

  sieve_word = *bit_array;
  k_bit_base = threadIdx.x * words_per_thread * 32;
  for (i = total_bit_count - bitcount[threadIdx.x]; ; i++) {
    int bit_to_test;

// Make sure we have a non-zero sieve word

    while (sieve_word == 0) {
      if (--words_per_thread == 0) break;
      sieve_word = *++bit_array;
      k_bit_base += 32;
    }

// Check if this thread has processed all its set bits

    if (sieve_word == 0) break;

// Find a bit to test in the sieve word

    bit_to_test = 31 - ___clz (sieve_word);
    sieve_word &= ~(1 << bit_to_test);

// Copy the k value to the shared memory array

    smem[i] = k_bit_base + bit_to_test;
  }

  __syncthreads();

// Here, all warps in our block have placed their candidates in shared memory.
// Now we can start TFing candidates.

// Pre-init values used to compute factor

  // Compute base k value
  k_base.d0 = __add_cc (k_base.d0, __umul32  (blockIdx.x * bits_to_process, NUM_CLASSES));
  k_base.d1 = __addc   (k_base.d1, __umul32hi(blockIdx.x * bits_to_process, NUM_CLASSES)); /* k values are limited to 64 bits */

// Loop til the k values written to shared memory are exhausted

  for (i = threadIdx.x; i < total_bit_count; i += THREADS_PER_BLOCK) {
    int96 k;
    int160 a, u, tmp160;
    int320 b, tmp320;
    float ff;
    int k_delta, shifter;

// Get the (k - k_base) value to test

    k_delta = smem[i];

// Compute f = k * 2^exp + 1

    k.d0 = __add_cc (k_base.d0, k_delta * NUM_CLASSES);	// k
    k.d1 = __addc   (k_base.d1, 0);

    f.d0 = 1;						// f = k * 2^exp + 1
    f.d1 = 0;
    f.d2 = (k.d0 << (exp - 64));
    f.d3 = (k.d1 << (exp - 64)) + (k.d0 >> (32 - (exp - 64)));
    f.d4 =                        (k.d1 >> (32 - (exp - 64)));

/*
ff = f as float
Precalculated here since it is the same for all steps in the following loop */
    ff= __uint2float_rn(f.d4);
    ff= ff * 4294967296.0f + __uint2float_rn(f.d3);
    ff=__int_as_float(0x3f7ffffb) / ff;			// just a little bit below 1.0f so we always underestimate the quotient

    tmp320.d9 = 1 << (bits_max - 1);			// tmp320 = 2^(159 + bits_in_f)
    tmp320.d8 = 0; tmp320.d7 = 0; tmp320.d6 = 0; tmp320.d5 = 0; tmp320.d4 = 0;
    tmp320.d3 = 0; tmp320.d2 = 0; tmp320.d1 = 0; tmp320.d0 = 0;

    // Could write optimized div_320_160 with so many tmp320 elements known to be zero
    div_320_160(&u,tmp320,f,ff);			// u = floor(2^(159 + bits_in_f) / f), giving 160 bits of precision

							// b_preinit = 2^256
							// a = b_preinit / 2 ^ (bits_in_f - 1)
							// tmp320 = a * u = (b_preinit / 2 ^ (bits_in_f - 1)) * (2 ^ (159 + bits_in_f) / f)     (ignore the floor functions for now)
							// a = tmp320 / 2^160, which if we do the math simplifies to the quotient: b_preinit / f
    a.d0 = (u.d1 >> (bits_max - 1)) + (u.d2 << (32 - (bits_max - 1)));
    a.d1 = (u.d2 >> (bits_max - 1)) + (u.d3 << (32 - (bits_max - 1)));
    a.d2 = (u.d3 >> (bits_max - 1)) + (u.d4 << (32 - (bits_max - 1)));
    a.d3 = (u.d4 >> (bits_max - 1));

    mul_160_F64_95_initial_special(&tmp160, a, f);	// tmp160 = quotient * f, we only compute the low 160-bits here

    a.d0 = __sub_cc (0, tmp160.d0);			// Compute the remainder
    a.d1 = __subc_cc(0, tmp160.d1);			// we do not need the upper digits of b and tmp160 because the result is 0 after subtraction!
    a.d2 = __subc_cc(0, tmp160.d2);
    a.d3 = __subc_cc(0, tmp160.d3);
    a.d4 = __subc   (0, tmp160.d4);

    for (shifter = 0; shifter < exp - 2 - 8; shifter++)
    {
							// On input a is at most 155.459 bits (see end of this loop)

      square_160_320(&b, a);				// b = a^2, b is at most 310.918 bits

      tmp160.d0 = (b.d4 >> (bits_max - 1)) + (b.d5 << (32 - (bits_max - 1))); // a = b / (2 ^ (bits_in_f - 1)), a is at most 159.918 bits
      tmp160.d1 = (b.d5 >> (bits_max - 1)) + (b.d6 << (32 - (bits_max - 1)));
      tmp160.d2 = (b.d6 >> (bits_max - 1)) + (b.d7 << (32 - (bits_max - 1)));
      tmp160.d3 = (b.d7 >> (bits_max - 1)) + (b.d8 << (32 - (bits_max - 1)));
      tmp160.d4 = (b.d8 >> (bits_max - 1)) + (b.d9 << (32 - (bits_max - 1)));

      mul_160_320_no_low5(&a, tmp160, u);		// a = (b / 2 ^ (bits_in_f - 1)) * (2 ^ (159 + bits_in_f) / f) / 2^160   (ignore the floor functions for now)
							// which if we do the math simplifies to the quotient: b / f

							// The quotient is off by at most 10.  A full mul_160_320 would add 9 partial results
							// into tmp320.d4 which could have generated 8 carries into tmp320.d5.
							// Also, since u was generated with the floor function, it could be low by up to
							// almost 1.  If we account for this a value up to a.d4 could have been added into
							// tmp320.d4 possibly generating a carry.  Similarly, a was generated by a floor
							// function, and could thus be low by almost 1.  If we account for this a value up
							// to u.d4 could have been added into tmp320.d4 possibly generating a carry.
							// A grand total of up to 10 carries lost.

      mul_160_F64_95_special(&tmp160, a, f);		// tmp160 = quotient * f, we only compute the low 160-bits here

      a.d0 = __sub_cc (b.d0, tmp160.d0);		// Compute the remainder
      a.d1 = __subc_cc(b.d1, tmp160.d1);		// we do not need the upper digits of b and tmp160 because the result is 0 after subtraction!
      a.d2 = __subc_cc(b.d2, tmp160.d2);
      a.d3 = __subc_cc(b.d3, tmp160.d3);
      a.d4 = __subc   (b.d4, tmp160.d4);
							// Since the quotient was up to 10 too small, the remainder has a maximum value of 11*f,
							// or 152 bits + log2 (11) bits, which is 155.459 bits.							// f values up to 2^151.541.
    }

    mod_simple_160(&finalrem, a, f, ff);		// Adjustment.  The code above may produce an a that is too large by up to 10 times f.

#if 0
    if(cmp_ge_160(finalrem,f))
    {
      printf("EEEEEK, final rem is >= f\n");
    }
if (blockIdx.x == 12 && threadIdx.x == 4){
mul_160_320_no_low5(&tmp160,u,f);
printf ("    f: %08X%08X%08X%08X%08X\r\n", f.d4, f.d3, f.d2, f.d1, f.d0);
printf ("u    : %X %X %X %X %X\r\n", u.d4, u.d3, u.d2, u.d1, u.d0);
printf ("u * f: %X %X %X %X %X\r\n", tmp160.d4, tmp160.d3, tmp160.d2, tmp160.d1, tmp160.d0);
printf ("  rem: %08X%08X%08X%08X%08X\r\n", finalrem.d4, finalrem.d3, finalrem.d2, finalrem.d1, finalrem.d0);
}
#endif

/* check if we found a factor and write the factor to RES[] */

    if(finalrem.d1 == 0)
    if((finalrem.d3 == f.d3 && finalrem.d2 == f.d2 && finalrem.d0 == 0 && finalrem.d4 == f.d4) ||
       (finalrem.d3 == 0    && finalrem.d2 == 0    && finalrem.d0 == 1 && finalrem.d4 == 0))
    {
      int index=atomicInc(&RES[0],10000);
      if(index<10)				/* limit to 10 factors per class */
      {
        RES[index*6 + 1]=0;
	RES[index*6 + 2]=f.d4;
	RES[index*6 + 3]=f.d3;
	RES[index*6 + 4]=f.d2;
	RES[index*6 + 5]=0;
	RES[index*6 + 6]=1;
      }
    }
  }

/* finally write occasional result for validation by C code */

  if(blockIdx.x == 0 && threadIdx.x == 0)
  {
    RES[84]=0;
    RES[85]=f.d4;
    RES[86]=f.d3;
    RES[87]=f.d2;
    RES[88]=0;
    RES[89]=1;
    RES[94]=0;
    RES[95]=finalrem.d4;
    RES[96]=finalrem.d3;
    RES[97]=finalrem.d2;
    RES[98]=finalrem.d1;
    RES[99]=finalrem.d0;
  }
}


//
// Kernel to factor Fermat numbers with factors f = k*2^N+1, 64 <= N <= 95.  Works on f between 153 and 160 bits inclusive
//

__global__ void
__launch_bounds__(THREADS_PER_BLOCK, KERNEL_MIN_BLOCKS) mfaktc_barrett160_F64_95gs(unsigned int exp, int96 k_base, unsigned int *bit_array, unsigned int bits_to_process, unsigned int *RES)
{
  int i, words_per_thread, sieve_word, k_bit_base, total_bit_count;
  int160 f, finalrem;
  __shared__ volatile unsigned short bitcount[256];	// Each thread of our block puts bit-counts here
  extern __shared__ unsigned short smem[];		// Write bits to test here.  Launching program must estimate
							// how much shared memory to allocate based on number of primes sieved.

  // Get pointer to section of the bit_array this thread is processing.

  words_per_thread = bits_to_process / 8192;
  bit_array += blockIdx.x * bits_to_process / 32 + threadIdx.x * words_per_thread;

// Count number of bits set in this thread's word(s) from the bit_array

  bitcount[threadIdx.x] = 0;
  for (i = 0; i < words_per_thread; i++)
    bitcount[threadIdx.x] += ___popcnt(bit_array[i]);

// Create total count of bits set in block up to and including this threads popcnt.
// Kudos to Rocke Verser for the population counting code.
// CAUTION:  Following requires 256 threads per block

  // First five tallies remain within one warp.  Should be in lock-step.
  if (threadIdx.x & 1)        // If we are running on any thread 0bxxxxxxx1, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[threadIdx.x - 1];

  if (threadIdx.x & 2)        // If we are running on any thread 0bxxxxxx1x, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 2) | 1];

  if (threadIdx.x & 4)        // If we are running on any thread 0bxxxxx1xx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 4) | 3];

  if (threadIdx.x & 8)        // If we are running on any thread 0bxxxx1xxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 8) | 7];

  if (threadIdx.x & 16)       // If we are running on any thread 0bxxx1xxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 16) | 15];

  // Further tallies are across warps.  Must synchronize
  __syncthreads();
  if (threadIdx.x  & 32)      // If we are running on any thread 0bxx1xxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 32) | 31];

  __syncthreads();
  if (threadIdx.x & 64)       // If we are running on any thread 0bx1xxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[(threadIdx.x - 64) | 63];

  __syncthreads();
  if (threadIdx.x & 128)       // If we are running on any thread 0b1xxxxxxx, tally neighbor's count.
    bitcount[threadIdx.x] += bitcount[127];

  // At this point, bitcount[...] contains the total number of bits for the indexed
  // thread plus all lower-numbered threads.  I.e., bitcount[255] is the total count.

  __syncthreads();
  total_bit_count = bitcount[255];

//POSSIBLE OPTIMIZATION - bitcounts and smem could use the same memory space if we'd read bitcount into a register
// and sync threads before doing any writes to smem.

//POSSIBLE SANITY CHECK -- is there any way to test if total_bit_count exceeds the amount of shared memory allocated?

// Loop til this thread's section of the bit array is finished.

  sieve_word = *bit_array;
  k_bit_base = threadIdx.x * words_per_thread * 32;
  for (i = total_bit_count - bitcount[threadIdx.x]; ; i++) {
    int bit_to_test;

// Make sure we have a non-zero sieve word

    while (sieve_word == 0) {
      if (--words_per_thread == 0) break;
      sieve_word = *++bit_array;
      k_bit_base += 32;
    }

// Check if this thread has processed all its set bits

    if (sieve_word == 0) break;

// Find a bit to test in the sieve word

    bit_to_test = 31 - ___clz (sieve_word);
    sieve_word &= ~(1 << bit_to_test);

// Copy the k value to the shared memory array

    smem[i] = k_bit_base + bit_to_test;
  }

  __syncthreads();

// Here, all warps in our block have placed their candidates in shared memory.
// Now we can start TFing candidates.

// Pre-init values used to compute factor

  // Compute base k value
  k_base.d0 = __add_cc (k_base.d0, __umul32  (blockIdx.x * bits_to_process, NUM_CLASSES));
  k_base.d1 = __addc   (k_base.d1, __umul32hi(blockIdx.x * bits_to_process, NUM_CLASSES)); /* k values are limited to 64 bits */

// Loop til the k values written to shared memory are exhausted

  for (i = threadIdx.x; i < total_bit_count; i += THREADS_PER_BLOCK) {
    int96 k;
    int192 a, u, tmp192;
    int352 b, tmp352;
    float ff;
    int k_delta, shifter;

// Get the (k - k_base) value to test

    k_delta = smem[i];

// Compute f = k * 2^exp + 1

    k.d0 = __add_cc (k_base.d0, k_delta * NUM_CLASSES);	// k
    k.d1 = __addc   (k_base.d1, 0);

    f.d0 = 1;						// f = k * 2^exp + 1
    f.d1 = 0;
    f.d2 = (k.d0 << (exp - 64));
    f.d3 = (k.d1 << (exp - 64)) + (k.d0 >> (32 - (exp - 64)));
    f.d4 =                        (k.d1 >> (32 - (exp - 64)));

/*
ff = f as float
Precalculated here since it is the same for all steps in the following loop */
    ff= __uint2float_rn(f.d4);
    ff= ff * 4294967296.0f + __uint2float_rn(f.d3);
    ff=__int_as_float(0x3f7ffffb) / ff;			// just a little bit below 1.0f so we always underestimate the quotient

    tmp352.d10 = 0x00010000;				// tmp352 is 2^336
    tmp352.d9 = 0; tmp352.d8 = 0; tmp352.d7 = 0; tmp352.d6 = 0; tmp352.d5 = 0;
    tmp352.d4 = 0; tmp352.d3 = 0; tmp352.d2 = 0; tmp352.d1 = 0; tmp352.d0 = 0;

    // Could write optimized div_352_160 with so many tmp352 elements known to be zero
    div_352_160(&u,tmp352,f,ff);			// u = floor(2^336 / f).  This requires f >= 145 bits.

							// b_preinit = 2^256
							// a = b_preinit / 2^144 = 2^112
							// tmp352 = a * u = (b_preinit / 2^144) * (2^336 / f)     (ignore the floor functions for now)
    a.d0 = (u.d3 << 16) + (u.d2 >> 16);			// a = tmp352 / 2^192, which if we do the math simplifies to the quotient: b_preinit / f
    a.d1 = (u.d4 << 16) + (u.d3 >> 16);
    a.d2 = (u.d5 << 16) + (u.d4 >> 16);
    a.d3 =                (u.d5 >> 16);

    mul_192_160_F64_95_initial_special16(&tmp192, a, f); // tmp192 = quotient * f, we only compute the low 192-bits here

    a.d0 = __sub_cc (0, tmp192.d0);			// Compute the remainder
    a.d1 = __subc_cc(0, tmp192.d1);			// we do not need the upper digits of b_preinit and tmp192 because the result is 0 after subtraction!
    a.d2 = __subc_cc(0, tmp192.d2);
    a.d3 = __subc_cc(0, tmp192.d3);
    a.d4 = __subc_cc(0, tmp192.d4);
    a.d5 = __subc   (0, tmp192.d5);

    for (shifter = 0; shifter < exp - 2 - 8; shifter++)
    {
							// On input a is at most 163.700 bits (see end of this loop)

      square_192_352(&b, a);				// b = a^2, b is at most 327.400 bits

      tmp192.d0 = (b.d5 << 16) + (b.d4 >> 16);		// a = b / 2^144, a is at most 183.400 bits
      tmp192.d1 = (b.d6 << 16) + (b.d5 >> 16);
      tmp192.d2 = (b.d7 << 16) + (b.d6 >> 16);
      tmp192.d3 = (b.d8 << 16) + (b.d7 >> 16);
      tmp192.d4 = (b.d9 << 16) + (b.d8 >> 16);
      tmp192.d5 = (b.d10 << 16) + (b.d9 >> 16);

      mul_192_384_no_low6(&a, tmp192, u);		// a = (b / 2^144) * (2^336 / f) / 2^192    (ignore the floor functions for now)
							// which if we do the math simplifies to the quotient: b / f

							// The quotient is off by at most 12.  A full mul_192_384 would add 11 partial results
							// into tmp384.d5 which could have generated 10 carries into tmp384.d6.
							// Also, since u was generated with the floor function, it could be low by up to
							// almost 1.  If we account for this a value up to a.d5 could have been added into
							// tmp384.d5 possibly generating a carry.  Similarly, a was generated by a floor
							// function, and could thus be low by almost 1.  If we account for this a value up
							// to u.d5 could have been added into tmp384.d5 possibly generating a carry.
							// A grand total of up to 12 carries lost.

      mul_192_160_F64_95_special(&tmp192, a, f);	// tmp192 = quotient * f, we only compute the low 192-bits here

      a.d0 = __sub_cc (b.d0, tmp192.d0);		// Compute the remainder
      a.d1 = __subc_cc(b.d1, tmp192.d1);		// we do not need the upper digits of b and tmp192 because the result is 0 after subtraction!
      a.d2 = __subc_cc(b.d2, tmp192.d2);
      a.d3 = __subc_cc(b.d3, tmp192.d3);
      a.d4 = __subc_cc(b.d4, tmp192.d4);
      a.d5 = __subc   (b.d5, tmp192.d5);
							// Since the quotient was up to 12 too small, the remainder has a maximum value of 13*f,
							// or 160 bits + log2 (13) bits, which is 163.700 bits.
    }

    mod_simple_192_160(&finalrem, a, f, ff);		// Adjustment.  The code above may produce an a that is too large by up to 13 times f.

#if 0
    if(cmp_ge_160(finalrem,f))
    {
      printf("EEEEEK, final rem is >= f\n");
    }
if (blockIdx.x == 12 && threadIdx.x == 4){
int192 f192;	    
f192.d0 = f.d0;
f192.d1 = f.d1;
f192.d2 = f.d2;
f192.d3 = f.d3;
f192.d4 = f.d4;
f192.d5 = 0;
mul_192_384_no_low6(&tmp192,u,f192);
printf ("    f: %08X%08X%08X%08X%08X\r\n", f.d4, f.d3, f.d2, f.d1, f.d0);
printf ("u    : %X %X %X %X %X %X\r\n", u.d5, u.d4, u.d3, u.d2, u.d1, u.d0);
printf ("u * f: %X %X %X %X %X %X\r\n", tmp192.d5, tmp192.d4, tmp192.d3, tmp192.d2, tmp192.d1, tmp192.d0);
printf ("  rem: %08X%08X%08X%08X%08X\r\n", finalrem.d4, finalrem.d3, finalrem.d2, finalrem.d1, finalrem.d0);
}
#endif

/* check if we found a factor and write the factor to RES[] */

    if(finalrem.d1 == 0)
    if((finalrem.d3 == f.d3 && finalrem.d2 == f.d2 && finalrem.d0 == 0 && finalrem.d4 == f.d4) ||
       (finalrem.d3 == 0    && finalrem.d2 == 0    && finalrem.d0 == 1 && finalrem.d4 == 0))
    {
      int index=atomicInc(&RES[0],10000);
      if(index<10)				/* limit to 10 factors per class */
      {
        RES[index*6 + 1]=0;
	RES[index*6 + 2]=f.d4;
	RES[index*6 + 3]=f.d3;
	RES[index*6 + 4]=f.d2;
	RES[index*6 + 5]=0;
	RES[index*6 + 6]=1;
      }
    }
  }

/* finally write occasional result for validation by C code */

  if(blockIdx.x == 0 && threadIdx.x == 0)
  {
    RES[84]=0;
    RES[85]=f.d4;
    RES[86]=f.d3;
    RES[87]=f.d2;
    RES[88]=0;
    RES[89]=1;
    RES[94]=0;
    RES[95]=finalrem.d4;
    RES[96]=finalrem.d3;
    RES[97]=finalrem.d2;
    RES[98]=finalrem.d1;
    RES[99]=finalrem.d0;
  }
}
