/*
This file is part of mfaktc.
Copyright (C) 2009, 2010, 2012  Oliver Weihe (o.weihe@t-online.de)

mfaktc is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mfaktc is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                
You should have received a copy of the GNU General Public License
along with mfaktc.  If not, see <http://www.gnu.org/licenses/>.
*/

// Welcome to the "dumb num" routines

// remove leading zeroes from a
void dn_normalize (unsigned int *a)
{
  while (a[0] && a[a[0]] == 0) a[0]--;
}

// copy a to b
void dn_copy (unsigned int *a, unsigned int *b)
{
  unsigned int i;

  dn_normalize(a);
  for (i = 0; i <= a[0]; i++) b[i] = a[i];
}

// return TRUE if a = b
int dn_equal (unsigned int *a, unsigned int *b)
{
  unsigned int i;

  dn_normalize(a);
  dn_normalize(b);
  if (a[0] != b[0]) return 0;
  for (i = a[0]; i >= 1; i--)
    if (a[i] != b[i]) return 0;
  return 1;
}

// return TRUE if a >= b
int dn_greater_or_equal (unsigned int *a, unsigned int *b)
{
  unsigned int i;

  dn_normalize(a);
  dn_normalize(b);
  if (a[0] > b[0]) return 1;
  if (a[0] < b[0]) return 0;
  for (i = a[0]; i >= 1; i--) {
    if (a[i] > b[i]) return 1;
    if (a[i] < b[i]) return 0;
  }
  return 1;
}

// return c = a + b
void dn_add (unsigned int *a, unsigned int *b, unsigned int *c)
{
  unsigned int tmp1[64], tmp2[64];
  unsigned int carry = 0;
  unsigned int i;

  dn_normalize(a);
  dn_normalize(b);
  if (a == c) { dn_copy(a, tmp1); a = tmp1; }
  if (b == c) { dn_copy(b, tmp2); b = tmp2; }
  for (i = 1; i <= a[0] || i <= b[0]; i++) {
    unsigned long long tmp = (unsigned long long) carry;
    if (i <= a[0]) tmp += (unsigned long long) a[i];
    if (i <= b[0]) tmp += (unsigned long long) b[i];
    carry = (unsigned int) (tmp >> 32);
    c[i] = (unsigned int) (tmp & 0xFFFFFFFF);
  }
  c[0] = i - 1;
}

// return c = a - b
void dn_sub (unsigned int *a, unsigned int *b, unsigned int *c)
{
  unsigned int tmp1[64], tmp2[64];
  unsigned int i;
  unsigned int borrow = 0;

  dn_normalize(a);
  dn_normalize(b);
  if (a == c) { dn_copy(a, tmp1); a = tmp1; }
  if (b == c) { dn_copy(b, tmp2); b = tmp2; }
  for (i = 1; i <= a[0]; i++) {
    unsigned long long tmp;
    tmp = (unsigned long long) a[i] - (unsigned long long) borrow;
    if (i <= b[0]) tmp -= (unsigned long long) b[i];
    borrow = (unsigned int) (tmp >> 32) & 1;
    c[i] = (unsigned int) (tmp & 0xFFFFFFFF);
  }
  c[0] = a[0];
//if (c[c[0]] > a[a[0]]) printf ("sub failure\n");
}

// decrement the most significant word of a -- specialized routine used in division
void dn_dec_top_word (unsigned int *a)
{
  if (a[a[0]] == 1) {
    a[0]--;
    a[a[0]] = 0xFFFFFFFF;
  } else
    a[a[0]]--;
}

// return c = a * b
void dn_mul (unsigned int *a, unsigned int *b, unsigned int *c)
{
  unsigned int tmp1[64], tmp2[64];
  unsigned int i, j;

  dn_normalize(a);
  dn_normalize(b);
  if (a == c) { dn_copy(a, tmp1); a = tmp1; }
  if (b == c) { dn_copy(b, tmp2); b = tmp2; }
  for (i = 1; i <= a[0] + b[0]; i++) c[i] = 0;
  for (i = 1; i <= a[0]; i++) {
    unsigned int carry = 0;
    for (j = 1; j <= b[0]; j++) {
      unsigned long long tmp;
      tmp = (unsigned long long) a[i] * (unsigned long long) b[j] + (unsigned long long) carry + (unsigned long long) c[i+j-1];
      carry = (unsigned int) (tmp >> 32);
      c[i+j-1] = (unsigned int) (tmp & 0xFFFFFFFF);
    }
    c[i+j-1] = carry;
  }
  c[0] = a[0] + b[0];
}

// return b = a^2
void dn_square (unsigned int *a, unsigned int *b)
{
  dn_mul (a, a, b);
}

// return c = a % b
void dn_mod (unsigned int *a, unsigned int *b, unsigned int *c)
{
  unsigned int tmp2[64];
  double top_c, top_b;

  dn_normalize(a);
  dn_normalize(b);
  if (b == c) { dn_copy(b, tmp2); b = tmp2; }

  dn_copy (a, c);
  top_b = (double) b[b[0]] * 4294967296.0;
  if (b[0] > 1) top_b += (double) b[b[0] - 1];
  while (dn_greater_or_equal (c, b)) {
    unsigned int quotient[64], tmp[64];
    unsigned int i;
    top_c = (double) c[c[0]] * 4294967296.0;
    if (c[0] > 1) top_c += (double) c[c[0] - 1];
    if (top_c >= top_b || c[0] == b[0]) {
      quotient[0] = c[0] - b[0] + 1;
      for (i = 1; i < quotient[0]; i++) quotient[i] = 0;
      quotient[i] = (unsigned int) floor (top_c / top_b);
    } else {
      quotient[0] = c[0] - b[0];
      for (i = 1; i < quotient[0]; i++) quotient[i] = 0;
      quotient[i] = (unsigned int) floor (top_c * 4294967296.0 / top_b);
    }
    if (quotient[i] == 0) quotient[i] = 1;
    for ( ; ; ) {
      dn_mul (b, quotient, tmp);
      if (dn_greater_or_equal (c, tmp)) break;
      dn_dec_top_word (quotient);
    }
//printf ("quot: %u, %u, c: %u %u %u, b: %u %u %u\n", quotient[0], quotient[i], c[0], c[c[0]], c[c[0]-1], b[0], b[b[0]], b[b[0]-1]);
    dn_sub (c, tmp, c);
  }
}

// return a mod b
unsigned int dn_mod_int (unsigned int *a, unsigned int b)
{
  unsigned int i;
  unsigned long long rem;

  dn_normalize(a);
  for (i = a[0], rem = 0; i >= 1; i--)
    rem = ((rem << 32) + a[i]) % b;
  return (unsigned int) rem;
}

// Validate a Fermat or double-Mersenne exponentiation

void validate_exponentiation(mystuff_t *mystuff, unsigned long long k_remaining)
{
  unsigned int fac[32], rem[32], tmp[64];
  unsigned int i;

  // get the data

  cudaMemcpy(mystuff->h_RES+80, mystuff->d_RES+80, 20*sizeof(int), cudaMemcpyDeviceToHost);
  fac[0] = 6;
  fac[1] = mystuff->h_RES[89];
  fac[2] = mystuff->h_RES[88];
  fac[3] = mystuff->h_RES[87];
  fac[4] = mystuff->h_RES[86];
  fac[5] = mystuff->h_RES[85];
  fac[6] = mystuff->h_RES[84];
  rem[0] = 6;
  rem[1] = mystuff->h_RES[99];
  rem[2] = mystuff->h_RES[98];
  rem[3] = mystuff->h_RES[97];
  rem[4] = mystuff->h_RES[96];
  rem[5] = mystuff->h_RES[95];
  rem[6] = mystuff->h_RES[94];

  // Kernels return a remainder of 0 when they test no factors.  This is just a happenstance
  // as it seems kernels zero out the uninitialized remainder held in registers.  To guard against
  // a kernel erroneously always returning zero, only bypass the validation if it is plausible that
  // the sieve zeroed all the k values.
  dn_normalize (rem);
  if (rem [0] == 0 && k_remaining < 16) return;  // Assume the sieve cleared all the k values

  // output the data

  if (mystuff->verbosity >= 3) {
    int192 factor, remainder;
    char factor_string[80], remainder_string[80];

    factor.d5=mystuff->h_RES[84];
    factor.d4=mystuff->h_RES[85];
    factor.d3=mystuff->h_RES[86];
    factor.d2=mystuff->h_RES[87];
    factor.d1=mystuff->h_RES[88];
    factor.d0=mystuff->h_RES[89];
    print_dez192(factor,factor_string);

    remainder.d5=mystuff->h_RES[94];
    remainder.d4=mystuff->h_RES[95];
    remainder.d3=mystuff->h_RES[96];
    remainder.d2=mystuff->h_RES[97];
    remainder.d1=mystuff->h_RES[98];
    remainder.d0=mystuff->h_RES[99];
    print_dez192(remainder,remainder_string);

    printf("Verifying (2^(2^%d)) %% %s = %s\n", (int) mystuff->exponent, factor_string, remainder_string);
  }

  // make sure the factor has no really small factors - this would indicate calculating the factor or GPU sieving or class_needed is broken

  if (dn_mod_int (fac, 2) == 0 ||
      dn_mod_int (fac, 3) == 0 ||
      dn_mod_int (fac, 5) == 0 ||
      dn_mod_int (fac, 7) == 0 ||
      dn_mod_int (fac, 11) == 0)
    printf ("ERROR: Class problems.  Factor divisible by 2, 3, 5, 7, or 11\n"), exit(1);
  for (i = 13; i <= 251; i++) {
    if (dn_mod_int (fac, i) == 0)
      printf ("ERROR: GPU sieve problems.  Factor divisible by %i\n", i), exit(1);
  }
  if (mystuff->verbosity >= 3) {
    for (i = 257; i <= 1999; i++) {
      if (dn_mod_int (fac, i) == 0)
        printf ("WARNING: Factor divisible by %i.  Only occasionally should GPU sieve let small factors slip through\n", i);
    }
  }

  // validate the exponentiation

  tmp[0] = 1; tmp[1] = 65536;
  for (i = 4; i < (mystuff->fermat_factoring ? mystuff->exponent - 2 : mystuff->exponent); i++) {
	  dn_square (tmp, tmp);
	  dn_mod (tmp, fac, tmp);
  }
  if (! dn_equal (tmp, rem))
      printf ("ERROR: Exponentiation failure\n"), exit(1);
}


// Figure out which Fermat number a factor divides

int which_fermat_number(mystuff_t *mystuff, int factor_index)
{
  unsigned int fac[32], tmp[64];
  unsigned int i, exp;

  // get the data

  fac[0] = 6;
  fac[1] = mystuff->h_RES[factor_index*6 + 6];
  fac[2] = mystuff->h_RES[factor_index*6 + 5];
  fac[3] = mystuff->h_RES[factor_index*6 + 4];
  fac[4] = mystuff->h_RES[factor_index*6 + 3];
  fac[5] = mystuff->h_RES[factor_index*6 + 2];
  fac[6] = mystuff->h_RES[factor_index*6 + 1];

  // Look for the exponentiation that returns fac-1 rather than 1

  for (exp = mystuff->exponent-2; exp > 5; exp--)
  {
    tmp[0] = 1; tmp[1] = 65536;
    for (i = 4; i < exp; i++) {
	  dn_square (tmp, tmp);
	  dn_mod (tmp, fac, tmp);
    }
    if (tmp[0] != 1 || tmp[1] != 1) break;
  }
  return (exp);
}
